#!/usr/bin/python
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
# GPL 2010
from __future__ import division, with_statement

import fractions
import os
import sys
import subprocess
import time

import ox


class AspectRatio(fractions.Fraction):
    def __new__(cls, numerator, denominator=None):
        if not denominator:
            ratio = map(int, numerator.split(':'))
            if len(ratio) == 1: ratio.append(1)
            numerator = ratio[0]
            denominator = ratio[1]
            #if its close enough to the common aspect ratios rather use that
            if abs(numerator/denominator - 4/3) < 0.03:
                numerator = 4
                denominator = 3
            elif abs(numerator/denominator - 16/9) < 0.02:
                numerator = 16
                denominator = 9
        return super(AspectRatio, cls).__new__(cls, numerator, denominator)

    @property
    def ratio(self):
        return "%d:%d" % (self.numerator, self.denominator)

def avinfo(filename, cached=True):
    if os.path.getsize(filename):
        info = ox.avinfo(filename, cached=cached)
        if 'video' in info and info['video'] and 'width' in info['video'][0]:
            if not 'display_aspect_ratio' in info['video'][0]:
                dar = AspectRatio(info['video'][0]['width'], info['video'][0]['height'])
                info['video'][0]['display_aspect_ratio'] = dar.ratio
        del info['path']
        if os.path.splitext(filename)[-1] in ('.srt', '.sub', '.idx', '.rar') and 'error' in info:
            del info['error']
        if 'code' in info and info['code'] == 'badfile':
            del info['code']
        return info
    return {'path': filename, 'size': 0}

def hash_prefix(h):
    return [h[:2], h[2:4], h[4:6], h[6:]]

def run_command(cmd, timeout=25):
    #print cmd
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
    while timeout > 0:
        time.sleep(0.2)
        timeout -= 0.2
        if p.poll() != None:
            return p.returncode
    if p.poll() == None:
        os.kill(p.pid, 9)
        killedpid, stat = os.waitpid(p.pid, os.WNOHANG)
    return p.returncode

def video_frame_positions(duration):
    pos = duration / 2
    #return [pos/4, pos/2, pos/2+pos/4, pos, pos+pos/2, pos+pos/2+pos/4]
    return map(int, [pos/2, pos, pos+pos/2])

def cleanup_tree(root):
    again = False
    for prefix, folders, files in os.walk(root):
        removed = []
        for f in files:
            if f in ('.DS_Store', 'Thumbs.db'):
                os.unlink(os.path.join(prefix, f))
                removed.append(f)
        files = list(set(files) - set(removed))
        if not folders and not files and len(prefix) > len(root):
            os.rmdir(prefix)
            again = True
    if again:
        cleanup_tree(root)


def basedir():
    base = os.path.join(os.path.expanduser('~'), '.ox')
    if sys.platform.startswith('win'):
        base = os.path.join(sys.environ['APPDATA'], 'pandora_client')
    return base



# encoding: utf-8
# vi:si:et:sw=4:sts=4:ts=4
import os
import json
import shutil
import time
import thread
from Queue import Queue
from threading import Thread

import ox
from twisted.web.resource import Resource
from twisted.web.static import File
from twisted.web.server import Site
from twisted.internet import reactor

import extract
from utils import hash_prefix

class UploadThread(Thread):
    def __init__(self, server):
        Thread.__init__(self)
        self.server = server

    def run(self):
        while True:
            oshash = self.server.upload.get()
            print oshash
            try:
                self.server.client.upload([oshash])
            except:
                print 'failed to upload', oshash
            self.server.upload.task_done()

class Server(Resource):

    def __init__(self, client):
        self.upload = Queue()
        self.client = client
        Resource.__init__(self)
        t = UploadThread(self)
        t.setDaemon(True)
        t.start()

    def active_encodes(self):
        conn, c = self.client._conn()
        site = self.client._config['url']
        active = int(time.mktime(time.localtime())) - 120
        status = 'active'
        sql = 'SELECT oshash FROM encode WHERE site = ? AND status = ? AND modified > ?'
        args = [site, status, active]
        c.execute(sql, tuple(args))
        files = [row[0] for row in c]
        #reset inactive encodes
        sql = 'UPDATE encode SET status = ? WHERE site = ? AND status = ? AND modified < ?'
        c.execute(sql, ('', site, 'active', active))
        conn.commit()
        return files

    def is_available(self, oshash):
        info = self.client.info(oshash)
        if info and not 'error' in info:
            for path in self.client.path(oshash):
                if os.path.exists(path):
                    return True
        return False

    def queued(self):
        site = self.client._config['url']
        files = self.client.get_encodes(site)
        available = filter(lambda oshash: self.is_available(oshash), files)
        unavailable = list(set(files) - set(available))
        return available, unavailable

    def queued_encodes(self):
        available, unavailable = self.queued()
        return available

    def update_status(self, oshash, status):
        conn, c = self.client._conn()
        site = self.client._config['url']
        modified = int(time.mktime(time.localtime()))
        c.execute(u'UPDATE encode SET status = ?, modified = ? WHERE site = ? AND oshash = ?', (status, modified, site, oshash))
        conn.commit()

    def media_path(self, oshash):
        return os.path.join(
            self.client.media_cache(),
            os.path.join(*hash_prefix(oshash)),
            self.client.profile(self.client.info(oshash))
        )

    def render_json(self, request, response):
        request.headers['Content-Type'] = 'application/json'
        return json.dumps(response, indent=2)

    def getChild(self, name, request):
        #make source media available via oshash
        if request.path.startswith('/get/'):
            oshash = request.path.split('/')[-1]
            for path in self.client.path(oshash):
                if os.path.exists(path):
                    f = File(path, 'application/octet-stream')
                    f.isLeaf = True
                    return f
        return self

    def render_PUT(self, request):
        if request.path.startswith('/upload'):
            parts = request.path.split('/')
            oshash = parts[-1]
            if len(oshash) == 16:
                path = self.media_path(oshash)
                tmp = path + '.tmp.webm'
                ox.makedirs(os.path.dirname(path))
                with open(tmp, 'wb') as f:
                    shutil.copyfileobj(request.content, f)
                shutil.move(tmp, path)
                self.update_status(oshash, 'done')
                self.upload.put(oshash)
                return self.render_json(request, {
                    'path': path
                })
        request.setResponseCode(404)
        return '404 unkown location'

    def render_POST(self, request):
        if request.path.startswith('/status'):
            oshash = request.path.split('/')[-1]
            error = request.args['error']
            self.update_status(oshash, 'failed')
            return self.render_json(request, {})
        request.setResponseCode(404)
        return '404 unkown location'

    def render_GET(self, request):
        if request.path.startswith('/next'):
            response = {}
            files = self.queued_encodes()
            for oshash in files:
                info = self.client.info(oshash)
                if not info or 'error' in info:
                    continue
                path = self.media_path(oshash)
                if os.path.exists(path) and os.stat(path).st_size > 0:
                    self.update_status(oshash, 'done')
                    self.upload.put(oshash)
                    continue
                for f in self.client.path(oshash):
                    if os.path.exists(f):
                        response['oshash'] = oshash
                        url = 'http://%s:%s/get/%s' % (request.host.host, request.host.port, oshash)
                        output = '/tmp/%s.%s' % (oshash, self.client.profile(info))
                        response['cmd'] = extract.video_cmd(url, output, self.client.profile(info), info)
                        response['cmd'][0] = 'ffmpeg'
                        response['output'] = output
                        self.update_status(oshash, 'active')
                        print oshash, f
                        return self.render_json(request, response)
            return self.render_json(request, response)
        elif request.path.startswith('/ping/'):
            parts = request.path.split('/')
            #FIXME: store client id somewhere
            client = parts[-1]
            oshash = parts[-2]
            self.update_status(oshash, 'active')
            return self.render_json(request, {})
        elif request.path.startswith('/update'):
            thread.start_new_thread(self.update, ())
            return self.render_json(request, {'status': True})
        elif request.path.startswith('/status'):
            queued, offline = self.queued()
            return self.render_json(request, {
                'active': self.active_encodes(),
                'queue': queued,
                'offline': offline
            })
        request.headers['Content-Type'] = 'text/html'
        data = 'pandora_client distributed encoding server'
        return data

    def update(self):
        self.client.scan([])
        self.client.sync([])
        self.client.update_encodes(True)

def run(client):
    root = Server(client)
    site = Site(root)
    port = 8789
    interface = '0.0.0.0'
    reactor.listenTCP(port, site, interface=interface)
    print 'listening on http://%s:%s' % (interface, port)
    client.update_encodes()
    reactor.run()
